-module(gen_emm).

-compile(inline).
-compile({inline_size, 128}).

-include_lib("kernel/include/logger.hrl").

-export([
   %% API for gen_emm
   start/0, start/1, start/2, start_link/0, start_link/1, start_link/2
   , start_monitor/0, start_monitor/1, start_monitor/2
   , stop/1, stop/3
   , call/3, call/4
   , send/3
   , send_request/3, wait_response/2, check_response/2
   , info_notify/2, call_notify/2
   , add_epm/3, add_sup_epm/3, del_epm/3
   , swap_epm/3, swap_sup_epm/3
   , which_epm/1

   %% gen callbacks
   , init_it/6

   %% sys callbacks
   , system_continue/3
   , system_terminate/4
   , system_code_change/4
   , system_get_state/1
   , system_replace_state/2
   , format_status/2

   %% Internal callbacks
   , wakeupFromHib/5

   %% logger callback
   , format_log/1, format_log/2
]).

-define(STACKTRACE(), element(2, erlang:process_info(self(), current_stacktrace))).

%% debug 调试相关宏定义
-define(NOT_DEBUG, []).
-define(SYS_DEBUG(Debug, Name, Msg),
   case Debug of
      ?NOT_DEBUG ->
         Debug;
      _ ->
         sys:handle_debug(Debug, fun print_event/3, Name, Msg)
   end).

-export_type([handler/0, handler_args/0, add_handler_ret/0,
   del_handler_ret/0]).

-type epmHandler() ::
   atom() |
   {atom(), term()}.

-type terminateArgs() ::
   term() |
   stop |
   removeEpm |
   {error, term()} |
   {stop, Reason :: term()} |
   {error, {'EXIT', Reason :: term()}}.

-type handler() ::
   atom() |
   {atom(), term()}.

-type handler_args() ::
   term().

-type add_handler_ret() ::
   ok |
   term() |
   {'EXIT', term()}.
-type del_handler_ret() ::
   ok |
   term() |
   {'EXIT', term()}.

-type serverName() ::
   {'local', atom()} |
   {'global', term()} |
   {'via', atom(), term()}.

-type serverRef() ::
   pid() |
   (LocalName :: atom())|
   {Name :: atom(), Node :: atom()} |
   {'global', term()} |
   {'via', atom(), term()}.

-type debug_flag() ::
   'trace' |
   'log' |
   'statistics' |
   'debug'|
   {'logfile', string()}.

-type startOpt() ::
   {'timeout', timeout()} |
   {'debug', [debug_flag()]} |
   {'spawn_opt', [proc_lib:start_spawn_option()]} |
   {'hibernate_after', timeout()}.

-type startRet() ::
   {'ok', pid()} |
   {'ok', {pid(), reference()}} |
   {'error', term()}.

-type requestId() :: term().

-record(epmHer, {
   epmId = undefined :: term(),
   epmM :: atom(),
   epmSup = undefined :: 'undefined' | pid(),
   epmS :: term()
}).

-callback init(InitArgs :: term()) ->
   {ok, State :: term()} |
   {ok, State :: term(), hibernate} |
   {error, Reason :: term()}.

-callback handleEvent(Event :: term(), State :: term()) ->
   kpS |
   removeEpm |
   {ok, NewState :: term()} |
   {ok, NewState :: term(), hibernate} |
   {swapEpm, NewState :: term(), Args1 :: term(), NewHandler :: epmHandler(), Args2 :: term()}.

-callback handleCall(Request :: term(), State :: term()) ->
   {removeEpm, Reply :: term()} |
   {reply, Reply :: term()} |
   {reply, Reply :: term(), NewState :: term()} |
   {reply, Reply :: term(), NewState :: term(), hibernate} |
   {swapEpm, Reply :: term(), NewState :: term(), Args1 :: term(), NewHandler :: epmHandler(), Args2 :: term()}.

-callback handleInfo(Info :: term(), State :: term()) ->
   kpS |
   removeEpm |
   {ok, NewState :: term()} |
   {ok, NewState :: term(), hibernate} |
   {swapEpm, NewState :: term(), Args1 :: term(), NewHandler :: epmHandler(), Args2 :: term()}.

-callback terminate(Args :: terminateArgs(), State :: term()) ->
   term().

-callback code_change(OldVsn :: (term() | {down, term()}), State :: term(), Extra :: term()) ->
   {ok, NewState :: term()}.

-callback format_status(Opt, StatusData) -> Status when
   Opt :: 'normal' | 'terminate',
   StatusData :: [PDict | State],
   PDict :: [{Key :: term(), Value :: term()}],
   State :: term(),
   Status :: term().

-optional_callbacks([
   handleInfo/2
   , terminate/2
   , code_change/3
   , format_status/2
]).

-spec start() -> startRet().
start() ->
   gen:start(?MODULE, nolink, ?MODULE, [], []).

-spec start(ServerName :: serverName() | [startOpt()]) -> startRet().
start(ServerName) when is_tuple(ServerName) ->
   gen:start(?MODULE, nolink, ServerName, ?MODULE, [], []);
start(Opts) when is_list(Opts) ->
   gen:start(?MODULE, nolink, ?MODULE, [], Opts).

-spec start(ServerName :: serverName(), Opts :: [startOpt()]) -> startRet().
start(ServerName, Opts) ->
   gen:start(?MODULE, nolink, ServerName, ?MODULE, [], Opts).

-spec start_link() -> startRet().
start_link() ->
   gen:start(?MODULE, link, ?MODULE, [], []).

-spec start_link(ServerName :: serverName() | [startOpt()]) -> startRet().
start_link(ServerName) when is_tuple(ServerName) ->
   gen:start(?MODULE, link, ServerName, ?MODULE, [], []);
start_link(Opts) when is_list(Opts) ->
   gen:start(?MODULE, link, ?MODULE, [], Opts).

-spec start_link(ServerName :: serverName(), Opts :: [startOpt()]) -> startRet().
start_link(ServerName, Opts) ->
   gen:start(?MODULE, link, ServerName, ?MODULE, [], Opts).

-spec start_monitor() -> startRet().
start_monitor() ->
   gen:start(?MODULE, monitor, ?MODULE, [], []).

-spec start_monitor(ServerName :: serverName() | [startOpt()]) -> startRet().
start_monitor(ServerName) when is_tuple(ServerName) ->
   gen:start(?MODULE, monitor, ServerName, ?MODULE, [], []);
start_monitor(Opts) when is_list(Opts) ->
   gen:start(?MODULE, monitor, ?MODULE, [], Opts).

-spec start_monitor(ServerName :: serverName(), Opts :: [startOpt()]) -> startRet().
start_monitor(ServerName, Opts) ->
   gen:start(?MODULE, monitor, ServerName, ?MODULE, [], Opts).

-spec stop(ServerRef :: serverRef()) -> 'ok'.
stop(ServerRef) ->
   gen:stop(ServerRef).

-spec stop(ServerRef :: serverRef(), Reason :: term(), Timeout :: timeout()) -> ok.
stop(ServerRef, Reason, Timeout) ->
   gen:stop(ServerRef, Reason, Timeout).

%% -spec init_it(pid(), 'self' | pid(), emgr_name(), module(), [term()], [_]) ->
init_it(Starter, self, ServerRef, Mod, Args, Options) ->
   init_it(Starter, self(), ServerRef, Mod, Args, Options);
init_it(Starter, Parent, ServerRef, _, _, Options) ->
   process_flag(trap_exit, true),
   Name = gen:name(ServerRef),
   Debug = gen:debug_options(Name, Options),
   HibernateAfterTimeout = gen:hibernate_after(Options),
   proc_lib:init_ack(Starter, {ok, self()}),
   receiveIng(Parent, Name, HibernateAfterTimeout, #{}, Debug, false).

-spec add_epm(serverRef(), epmHandler(), term()) -> term().
add_epm(EpmSrv, EpmHandler, Args) ->
   epmRpc(EpmSrv, {'$addEpm', EpmHandler, Args}).

-spec add_sup_epm(serverRef(), epmHandler(), term()) -> term().
add_sup_epm(EpmSrv, EpmHandler, Args) ->
   epmRpc(EpmSrv, {'$addSupEpm', EpmHandler, Args, self()}).

-spec del_epm(serverRef(), epmHandler(), term()) -> term().
del_epm(EpmSrv, EpmHandler, Args) ->
   epmRpc(EpmSrv, {'$deleteEpm', EpmHandler, Args}).

-spec swap_epm(serverRef(), {epmHandler(), term()}, {epmHandler(), term()}) -> 'ok' | {'error', term()}.
swap_epm(EpmSrv, {H1, A1}, {H2, A2}) ->
   epmRpc(EpmSrv, {'$swapEpm', H1, A1, H2, A2}).

-spec swap_sup_epm(serverRef(), {epmHandler(), term()}, {epmHandler(), term()}) -> 'ok' | {'error', term()}.
swap_sup_epm(EpmSrv, {H1, A1}, {H2, A2}) ->
   epmRpc(EpmSrv, {'$swapSupEpm', H1, A1, H2, A2, self()}).

-spec which_epm(serverRef()) -> [epmHandler()].
which_epm(EpmSrv) ->
   epmRpc(EpmSrv, '$which_handlers').

-spec info_notify(serverRef(), term()) -> 'ok'.
info_notify(EpmSrv, Event) ->
   epmRequest(EpmSrv, {'$gen_info', '$infoNotify', Event}).

-spec call_notify(serverRef(), term()) -> 'ok'.
call_notify(EpmSrv, Event) ->
   epmRpc(EpmSrv, {'$syncNotify', Event}).

-spec call(serverRef(), handler(), term()) -> term().
call(EpmSrv, EpmHandler, Query) ->
   epmRpc(EpmSrv, {'$epmCall', EpmHandler, Query}).

-spec call(serverRef(), epmHandler(), term(), timeout()) -> term().
call(EpmSrv, EpmHandler, Query, Timeout) ->
   epmRpc(EpmSrv, {'$epmCall', EpmHandler, Query}, Timeout).

send(EpmSrv, EpmHandler, Msg) ->
   epmRequest(EpmSrv, {'$gen_info', EpmHandler, Msg}).

-spec send_request(serverRef(), epmHandler(), term()) -> requestId().
send_request(EpmSrv, EpmHandler, Query) ->
   gen:send_request(EpmSrv, '$gen_call', {'$epmCall', EpmHandler, Query}).

-spec wait_response(RequestId :: requestId(), timeout()) ->
   {reply, Reply :: term()} | 'timeout' | {error, {Reason :: term(), serverRef()}}.
wait_response(RequestId, Timeout) ->
   case gen:wait_response(RequestId, Timeout) of
      {reply, {error, _} = Err} -> Err;
      Return -> Return
   end.

-spec check_response(Msg :: term(), RequestId :: requestId()) ->
   {reply, Reply :: term()} | 'no_reply' | {error, {Reason :: term(), serverRef()}}.
check_response(Msg, RequestId) ->
   case gen:check_response(Msg, RequestId) of
      {reply, {error, _} = Err} -> Err;
      Return -> Return
   end.

epmRpc(EpmSrv, Cmd) ->
   try gen:call(EpmSrv, '$gen_call', Cmd, infinity) of
      {ok, Reply} ->
         Reply
   catch Class:Reason ->
      erlang:raise(Class, {Reason, {?MODULE, call, [EpmSrv, Cmd, infinity]}}, ?STACKTRACE())
   end.

epmRpc(EpmSrv, Cmd, Timeout) ->
   try gen:call(EpmSrv, '$gen_call', Cmd, Timeout) of
      {ok, Reply} ->
         Reply
   catch Class:Reason ->
      erlang:raise(Class, {Reason, {?MODULE, call, [EpmSrv, Cmd, Timeout]}}, ?STACKTRACE())
   end.

epmRequest({global, Name}, Msg) ->
   try global:send(Name, Msg),
   ok
   catch _:_ -> ok
   end;
epmRequest({via, RegMod, Name}, Msg) ->
   try RegMod:send(Name, Msg),
   ok
   catch _:_ -> ok
   end;
epmRequest(EpmSrv, Cmd) ->
   EpmSrv ! Cmd,
   ok.

loopEntry(Parent, ServerName, HibernateAfterTimeout, EpmHers, Debug, true) ->
   proc_lib:hibernate(?MODULE, wakeupFromHib, [Parent, ServerName, HibernateAfterTimeout, EpmHers, Debug]);
loopEntry(Parent, ServerName,  HibernateAfterTimeout, EpmHers, Debug, _) ->
   receiveIng(Parent, ServerName, HibernateAfterTimeout, EpmHers, Debug, false).

wakeupFromHib(Parent, ServerName, HibernateAfterTimeout, EpmHers,  Debug) ->
   receiveIng(Parent, ServerName, HibernateAfterTimeout, EpmHers, Debug, true).

receiveIng(Parent, ServerName, HibernateAfterTimeout, EpmHers, Debug, IsHib) ->
   receive
      {system, From, Req} ->
         sys:handle_system_msg(Req, From, Parent, ?MODULE, Debug, [ServerName, HibernateAfterTimeout,  EpmHers, IsHib], IsHib);
      {'EXIT', Parent, Reason} ->
         terminate_server(Reason, Parent,  ServerName, EpmHers);
      {'$epm_call', From, Request} ->
         epmCallMsg(Parent, ServerName,  HibernateAfterTimeout, EpmHers, Debug, From, Request);
      {'$epm_info', CmdOrEmpHandler, Event} ->
         epmInfoMsg(Parent, ServerName,  HibernateAfterTimeout, EpmHers, Debug, CmdOrEmpHandler, Event);
      Msg ->
         handleMsg(Parent, ServerName,  HibernateAfterTimeout,EpmHers,  Debug, Msg)
   after HibernateAfterTimeout ->
      proc_lib:hibernate(?MODULE, wakeupFromHib, [Parent, ServerName, HibernateAfterTimeout, EpmHers, Debug])
   end.

epmCallMsg(Parent, ServerName, HibernateAfterTimeout, EpmHers, Debug, From, Request) ->
   NewDebug = ?SYS_DEBUG(Debug, ServerName, {call, From, Request}),
   case Request of
      '$which_handlers' ->
         reply(From, EpmHers),
         receiveIng(Parent, ServerName, HibernateAfterTimeout, EpmHers, NewDebug, false);
      {'$addEpm', EpmHandler, Args} ->
         {Reply, NewEpmHers, IsHib} = doAddEpm(EpmHers, EpmHandler, Args, undefined),
         reply(From, Reply),
         loopEntry(Parent, ServerName, HibernateAfterTimeout, NewEpmHers, NewDebug, IsHib);
      {'$addSupEpm', EpmHandler, Args, EpmSup} ->
         {Reply, NewEpmHers, IsHib} = doAddSupEpm(EpmHers, EpmHandler, Args, EpmSup),
         reply(From, Reply),
         loopEntry(Parent, ServerName, HibernateAfterTimeout, NewEpmHers, NewDebug, IsHib);
      {'$deleteEpm', EpmHandler, Args} ->
         {Reply, NewEpmHers} = doDeleteEpm(EpmHers, EpmHandler, Args),
         reply(From, Reply),
         receiveIng(Parent, ServerName, HibernateAfterTimeout, NewEpmHers, NewDebug, false);
      {'$swapEpm', EpmId1, Args1, EpmId2, Args2} ->
         {Reply, NewEpmHers, IsHib} = doSwapEpm(EpmHers, EpmId1, Args1, EpmId2, Args2),
         reply(From, Reply),
         loopEntry(Parent, ServerName, HibernateAfterTimeout, NewEpmHers, NewDebug, IsHib);
      {'$swapSupEpm', EpmId1, Args1, EpmId2, Args2, SupPid} ->
         {Reply, NewEpmHers, IsHib} = doSwapSupEpm(EpmHers, EpmId1, Args1, EpmId2, Args2, SupPid),
         reply(From, Reply),
         loopEntry(Parent, ServerName, HibernateAfterTimeout, NewEpmHers, NewDebug, IsHib);
      {'$syncNotify', Event} ->
         {NewEpmHers, IsHib} = doNotify(EpmHers, handleEvent, Event, false),
         reply(From, ok),
         loopEntry(Parent, ServerName, HibernateAfterTimeout, NewEpmHers, NewDebug, IsHib);
      {'$epmCall', EpmHandler, Query} ->
         {NewEpmHers, IsHib} = doEpmHandle(EpmHers, EpmHandler, handleCall, Query, From),
         loopEntry(Parent, ServerName, HibernateAfterTimeout, NewEpmHers, NewDebug, IsHib)
   end.

epmInfoMsg(Parent, ServerName, HibernateAfterTimeout, EpmHers, Debug, CmdOrEmpHandler, Event) ->
   NewDebug = ?SYS_DEBUG(Debug, ServerName, {info, CmdOrEmpHandler, Event}),
   case CmdOrEmpHandler of
      '$infoNotify' ->
         {NewEpmHers, IsHib} = doNotify(EpmHers, handleEvent, Event, false),
         loopEntry(Parent, ServerName, HibernateAfterTimeout, NewEpmHers, NewDebug, IsHib);
      EpmHandler ->
         {NewEpmHers, IsHib} = doEpmHandle(EpmHers, EpmHandler, handleInfo, Event,  false),
         loopEntry(Parent, ServerName, HibernateAfterTimeout, NewEpmHers, NewDebug, IsHib)
   end.

handleMsg(Parent, ServerName, HibernateAfterTimeout, EpmHers, Debug, Msg) ->
   NewDebug = ?SYS_DEBUG(Debug, ServerName, {in, Msg}),
   case Msg of
      {'EXIT', From, Reason} ->
         MSL1 = handle_exit(From, Reason, MSL, ServerName),
         loop(Parent, ServerName, MSL1, HibernateAfterTimeout, Debug, false);
      {_From, Tag, stop} ->
         catch terminate_server(normal, Parent, MSL, ServerName),
         reply(Tag, ok);
      {_From, Tag, get_modules} ->
         reply(Tag, get_modules(MSL)),
         loop(Parent, ServerName, MSL, HibernateAfterTimeout, Debug, false);
      _ ->
      {Hib, MSL1} = doNotify(EpmHers, handleInfo, Msg, false),
      loopEntry(Parent, ServerName, MSL1, HibernateAfterTimeout, NewDebug, Hib)
   end.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% EPM inner fun
addNewEpm(InitRet, EpmHers, Module, EpmId, EpmSup) ->
   case InitRet of
      {ok, State} ->
         EpmHer = #epmHer{epmId = EpmId, epmM = Module, epmS = State, epmSup = EpmSup},
         {ok, EpmHers#{EpmId => EpmHer}, false};
      {ok, State, hibernate} ->
         EpmHer = #epmHer{epmId = EpmId, epmM = Module, epmS = State, epmSup = EpmSup},
         {ok, EpmHers#{EpmId => EpmHer}, true};
      Other ->
         {Other, EpmHers, false}
   end.

doAddEpm(EpmHers, {Module, _SubId} = EpmId, Args, EpmSup) ->
   case EpmHers of
      #{EpmId := _EpmHer} ->
         {{error, existed}, EpmHers, false};
      _ ->
         try Module:init(Args) of
            Result ->
               addNewEpm(Result, EpmHers, Module, EpmId, EpmSup)
         catch
            throw:Ret ->
               addNewEpm(Ret, EpmHers, Module, EpmId, EpmSup);
            C:R ->
               {{error, {C, R, ?STACKTRACE()}}, EpmHers, false}
         end
   end;
doAddEpm(EpmHers, Module, Args, EpmSup) ->
   case EpmHers of
      #{Module := _EpmHer} ->
         {{error, existed}, EpmHers, false};
      _ ->
         try Module:init(Args) of
            Result ->
               addNewEpm(Result, EpmHers, Module, Module, EpmSup)
         catch
            throw:Ret ->
               addNewEpm(Ret, EpmHers, Module, Module, EpmSup);
            C:R ->
               {{error, {C, R, ?STACKTRACE()}}, EpmHers, false}
         end
   end.

doAddSupEpm(EpmHers, EpmHandler, Args, EpmSup) ->
   case doAddEpm(EpmHers, EpmHandler, Args, EpmSup) of
      {ok, _, _} = Result ->
         link(EpmSup),
         Result;
      Ret ->
         Ret
   end.

doSwapEpm(EpmHers, EpmId1, Args1, EpmMId, Args2) ->
   case EpmHers of
      #{EpmId1 := #epmHer{epmSup = EpmSup} = EpmHer} ->
         State2 = epmTerminate(EpmHer, Args1, swapped, {swapped, EpmMId, EpmSup}),
         NewEpmHers = maps:remove(EpmId1, EpmHers),
         case EpmSup of
            false ->
               doAddEpm(NewEpmHers, EpmMId, {Args2, State2}, undefined);
            _ ->
               doAddSupEpm(NewEpmHers, EpmMId, {Args2, State2}, EpmSup)
         end;
      undefined ->
         doAddEpm(EpmHers, EpmMId, {Args2, undefined}, undefined)
   end.

doSwapSupEpm(EpmHers, EpmId1, Args1, EpmMId, Args2, EpmSup) ->
   case EpmHers(EpmId1) of
      #{EpmId1 := EpmHer} ->
         State2 = epmTerminate(EpmHer, Args1, swapped, {swapped, EpmMId, EpmSup}),
         NewEpmHers = maps:remove(EpmId1, EpmHers),
         doAddSupEpm(NewEpmHers, EpmMId, {Args2, State2}, EpmSup);
      undefined ->
         doAddSupEpm(EpmHers, EpmMId, {Args2, undefined}, EpmSup)
   end.


doNotify(EpmHers, Func, Event, _Form) ->
   allNotify(maps:iterator(EpmHers), Func, Event, false, EpmHers, false).

allNotify(Iterator, Func, Event, From, TemEpmHers, IsHib) ->
   case maps:next(Iterator) of
      {K, _V, NextIterator} ->
         {NewEpmHers, NewIsHib} = doEpmHandle(TemEpmHers, K, Func, Event, From),
         allNotify(NextIterator, Func, Event, From, NewEpmHers, IsHib orelse NewIsHib);
      _ ->
         {TemEpmHers, IsHib}
   end.

doEpmHandle(EpmHers, EpmHandler, Func, Event, From) ->
   case EpmHers of
      #{EpmHandler := #epmHer{epmM = EpmM, epmS = EpmS} = EpmHer} ->
         try EpmM:Func(Event, EpmS) of
            Result ->
               handleEpmCR(Result, EpmHers, EpmHer, Event, From)
         catch
            throw:Ret ->
               handleEpmCR(Ret, EpmHers, EpmHer, Event, From);
            C:R ->
               epmTerminate(EpmHer, {error, {C, R, ?STACKTRACE()}}, Event, crash),
               NewEpmHers = maps:remove(EpmHandler, EpmHer),
               {NewEpmHers, false}
         end;
      _ ->
         {EpmHers, false}
   end.

doDeleteEpm(EpmHers, EpmHandler, Args) ->
   case EpmHers of
      #{EpmHandler := EpmHer} ->
         epmTerminate(EpmHer, Args, delete, normal),
         {ok, maps:remove(EpmHandler, EpmHers)};
      undefined ->
         {{error, module_not_found}, EpmHers}
   end.

%% handleEpmCallbackRet
handleEpmCR(Result, EpmHers, #epmHer{epmId = EpmId} = EpmHer, Event, From) ->
   case Result of
      kpS ->
         {EpmHers, false};
      {ok, NewEpmS} ->
         MewEpmHer = setelement(#epmHer.epmS, EpmHer, NewEpmS),
         {EpmHers#{EpmId := MewEpmHer}, false};
      {ok, NewEpmS, hibernate} ->
         MewEpmHer = setelement(#epmHer.epmS, EpmHer, NewEpmS),
         {EpmHers#{EpmId := MewEpmHer}, true};
      {swapEpm, NewEpmS, Args1, EpmMId, Args2} ->
         #epmHer{epmId = OldEpmMId, epmSup = EpmSup} = MewEpmHer = setelement(#epmHer.epmS, EpmHer, NewEpmS),
         State = epmTerminate(MewEpmHer, Args1, swapped, {swapped, OldEpmMId, EpmSup}),
         TemEpmHers = maps:remove(EpmId, EpmHers),
         {_, NewEpmHers, IsHib} =
            case EpmSup of
               undefined ->
                  doAddEpm(TemEpmHers, EpmMId, {Args2, State}, undefined);
               _ ->
                  doAddSupEpm(TemEpmHers, EpmMId, {Args2, State}, EpmSup)
            end,
         {NewEpmHers, IsHib};
      {swapEpm, Reply, NewEpmS, Args1, EpmMId, Args2} ->
         reply(From, Reply),
         #epmHer{epmId = OldEpmMId, epmSup = EpmSup} = MewEpmHer = setelement(#epmHer.epmS, EpmHer, NewEpmS),
         State = epmTerminate(MewEpmHer, Args1, swapped, {swapped, OldEpmMId, EpmSup}),
         TemEpmHers = maps:remove(EpmId, EpmHers),
         {_, NewEpmHers, IsHib} =
            case EpmSup of
               undefined ->
                  doAddEpm(TemEpmHers, EpmMId, {Args2, State}, undefined);
               _ ->
                  doAddSupEpm(TemEpmHers, EpmMId, {Args2, State}, EpmSup)
            end,
         {NewEpmHers, IsHib};
      removeEpm ->
         epmTerminate(EpmHer, removeEpm, remove, normal),
         {maps:remove(EpmId, EpmHers), false};
      {removeEpm, Reply} ->
         reply(From, Reply),
         epmTerminate(EpmHer, removeEpm, remove, normal),
         {maps:remove(EpmId, EpmHers), false};
      {reply, Reply} ->
         reply(From, Reply),
         {EpmHers, false};
      {reply, Reply, NewEpmS} ->
         reply(From, Reply),
         MewEpmHer = setelement(#epmHer.epmS, EpmHer, NewEpmS),
         {EpmHers#{EpmId := MewEpmHer}, false};
      {reply, Reply, NewEpmS, hibernate} ->
         reply(From, Reply),
         MewEpmHer = setelement(#epmHer.epmS, EpmHer, NewEpmS),
         {EpmHers#{EpmId := MewEpmHer}, true};
      Other ->
         epmTerminate(EpmHer, {error, Other}, Event, crash),
         {maps:remove(EpmId, EpmHers), false}
   end.

reportTerminate(EpmHer, crash, {error, Why}, LastIn, _) ->
   reportTerminate2(EpmHer, Why, LastIn);
%% How == normal | shutdown | {swapped, NewHandler, NewSupervisor}
reportTerminate(EpmHer, How, _, LastIn, _) ->
   reportTerminate2(EpmHer, How, LastIn).

reportTerminate2(#epmHer{epmSup = EpmSup, epmId = EpmId, epmS = State} = EpmHer, Reason, LastIn) ->
   report_error(EpmHer, Reason, State, LastIn),
   case EpmSup of
      undefined ->
         ok;
      _ ->
         EpmSup ! {gen_event_EXIT, EpmId, Reason},
         ok
   end.

report_error(_EpmHer, normal, _, _) -> ok;
report_error(_EpmHer, shutdown, _, _) -> ok;
report_error(_EpmHer, {swapped, _, _}, _, _) -> ok;
report_error(#epmHer{epmId = EpmId, epmM = EpmM}, Reason, State, LastIn) ->
   ?LOG_ERROR(
      #{
         label => {gen_emm, epm_terminate},
         handler => {EpmId, EpmM},
         name => undefined,
         last_message => LastIn,
         state => State,
         reason => Reason
      },
      #{
         domain => [otp],
         report_cb => fun gen_emm:epm_log/1,
         error_logger => #{tag => error}
      }).

epm_log(#{label := {gen_emm, epm_terminate}, handler := Handler, name := SName, last_message := LastIn, state := State, reason := Reason}) ->
   Reason1 =
      case Reason of
         {'EXIT', {undef, [{M, F, A, L} | MFAs]}} ->
            case code:is_loaded(M) of
               false ->
                  {'module could not be loaded', [{M, F, A, L} | MFAs]};
               _ ->
                  case erlang:function_exported(M, F, length(A)) of
                     true ->
                        {undef, [{M, F, A, L} | MFAs]};
                     false ->
                        {'function not exported', [{M, F, A, L} | MFAs]}
                  end
            end;
         {'EXIT', Why} ->
            Why;
         _ ->
            Reason
      end,
   {"** gen_emm emp handler ~p crashed.~n"
   "** Was installed in ~tp~n"
   "** Last event was: ~tp~n"
   "** When handler state == ~tp~n"
   "** Reason == ~tp~n", [Handler, SName, LastIn, State, Reason1]};
epm_log(#{label := {gen_emm, no_handle_info}, module := Module, message := Msg}) ->
   {"** Undefined handle_info in ~tp~n"
   "** Unhandled message: ~tp~n", [Module, Msg]}.

epmStopAll(EpmHers) ->
   allStop(maps:iterator(EpmHers)).

allStop(Iterator) ->
   case maps:next(Iterator) of
      {_K, V, NextIterator} ->
         epmTerminate(V, stop, 'receive', shutdown),
         case element(#epmHer.epmSup, V) of
            undefined ->
               ignore;
            EpmSup ->
               unlink(EpmSup)
         end,
         allStop(NextIterator);
      none ->
         ok
   end.

epmStopOne(ExitEmpSup, EpmHers) ->
   forStopOne(maps:iterator(EpmHers), ExitEmpSup, EpmHers).

forStopOne(Iterator, ExitEmpSup, TemEpmHers) ->
   case maps:next(Iterator) of
      {K, V, NextIterator} ->
         case element(#epmHer.epmSup, V) =:= ExitEmpSup of
            true ->
               epmTerminate(V, stop, 'receive', shutdown),
               forStopOne(NextIterator, ExitEmpSup, maps:remove(K, TemEpmHers));
            _ ->
               forStopOne(NextIterator, ExitEmpSup, TemEpmHers)
         end;
      none ->
         TemEpmHers
   end.

epmTerminate(#epmHer{epmM = EpmM, epmS = State} = EpmHer, Args, LastIn, Reason) ->
   case erlang:function_exported(EpmM, terminate, 2) of
      true ->
         Res = (catch EpmM:terminate(Args, State)),
         reportTerminate(EpmHer, Reason, Args, LastIn, Res),
         Res;
      false ->
         reportTerminate(EpmHer, Reason, Args, LastIn, ok),
         ok
   end.

reply({To, Ref}, Msg) ->
   To ! {Ref, Msg},
   ok.

terminate_server(Reason, Parent, ServerName, EpmHers) ->
   stop_handlers(EpmHers, ServerName),
   do_unlink(Parent, EpmHers),
   exit(Reason).



%% unlink the supervisor process of all supervised handlers.
%% We do not want a handler supervisor to EXIT due to the
%% termination of the event manager (server).
%% Do not unlink Parent !
do_unlink(Parent, MSL) ->
   lists:foreach(fun(Handler) when Handler#handler.supervised =:= Parent ->
      true;
      (Handler) when is_pid(Handler#handler.supervised) ->
         unlink(Handler#handler.supervised),
         true;
      (_) ->
         true
                 end,
      MSL).


%%-----------------------------------------------------------------
%% Callback functions for system messages handling.
%%-----------------------------------------------------------------
system_continue(Parent, Debug, [ServerName, HibernateAfterTimeout, EpmHers, Hib]) ->
   loopEntry(Parent, ServerName, HibernateAfterTimeout, EpmHers, Debug, Hib).

-spec system_terminate(_, _, _, [_]) -> no_return().
system_terminate(Reason, Parent, _Debug, [ServerName, _HibernateAfterTimeout, EpmHers, _Hib]) ->
   terminate_server(Reason, Parent, ServerName, EpmHers).

%%-----------------------------------------------------------------
%% Module here is sent in the system msg change_code.  It specifies
%% which module should be changed.
%%-----------------------------------------------------------------
system_code_change([ServerName, HibernateAfterTimeout, EpmHers, Hib], Module, OldVsn, Extra) ->
   MSL1 = lists:zf(fun(H) when H#handler.module =:= Module ->
      {ok, NewState} =
         Module:code_change(OldVsn,
            H#handler.state, Extra),
      {true, H#handler{state = NewState}};
      (_) -> true
                   end,
      MSL),
   {ok, [ServerName, MSL1, HibernateAfterTimeout, Hib]}.

system_get_state([_ServerName, MSL, _HibernateAfterTimeout, _Hib]) ->
   {ok, [{Mod, Id, State} || #handler{module = Mod, id = Id, state = State} <- MSL]}.

system_replace_state(StateFun, [ServerName, MSL, HibernateAfterTimeout, Hib]) ->
   {NMSL, NStates} =
      lists:unzip([begin
                      Cur = {Mod, Id, State},
                      try
                         NState = {Mod, Id, NS} = StateFun(Cur),
                         {HS#handler{state = NS}, NState}
                      catch
                         _:_ ->
                            {HS, Cur}
                      end
                   end || #handler{module = Mod, id = Id, state = State} = HS <- MSL]),
   {ok, NStates, [ServerName, NMSL, HibernateAfterTimeout, Hib]}.

%%-----------------------------------------------------------------
%% Format debug messages.  Print them as the call-back module sees
%% them, not as the real erlang messages.  Use trace for that.
%%-----------------------------------------------------------------
print_event(Dev, {in, Msg}, Name) ->
   case Msg of
      {notify, Event} ->
         io:format(Dev, "*DBG* ~tp got event ~tp~n", [Name, Event]);
      {_, _, {call, Handler, Query}} ->
         io:format(Dev, "*DBG* ~tp(~tp) got call ~tp~n",
            [Name, Handler, Query]);
      _ ->
         io:format(Dev, "*DBG* ~tp got ~tp~n", [Name, Msg])
   end;
print_event(Dev, Dbg, Name) ->
   io:format(Dev, "*DBG* ~tp : ~tp~n", [Name, Dbg]).

%% format_log/1 is the report callback used by Logger handler
%% error_logger only. It is kept for backwards compatibility with
%% legacy error_logger event handlers. This function must always
%% return {Format,Args} compatible with the arguments in this module's
%% calls to error_logger prior to OTP-21.0.
format_log(Report) ->
   Depth = error_logger:get_format_depth(),
   FormatOpts = #{chars_limit => unlimited,
      depth => Depth,
      single_line => false,
      encoding => utf8},
   format_log_multi(limit_report(Report, Depth), FormatOpts).

limit_report(Report, unlimited) ->
   Report;
limit_report(#{label := {gen_event, terminate},
   last_message := LastIn,
   state := State,
   reason := Reason} = Report,
   Depth) ->
   Report#{last_message => io_lib:limit_term(LastIn, Depth),
      state => io_lib:limit_term(State, Depth),
      reason => io_lib:limit_term(Reason, Depth)};
limit_report(#{label := {gen_event, no_handle_info},
   message := Msg} = Report,
   Depth) ->
   Report#{message => io_lib:limit_term(Msg, Depth)}.

%% format_log/2 is the report callback for any Logger handler, except
%% error_logger.
format_log(Report, FormatOpts0) ->
   Default = #{chars_limit => unlimited,
      depth => unlimited,
      single_line => false,
      encoding => utf8},
   FormatOpts = maps:merge(Default, FormatOpts0),
   IoOpts =
      case FormatOpts of
         #{chars_limit := unlimited} ->
            [];
         #{chars_limit := Limit} ->
            [{chars_limit, Limit}]
      end,
   {Format, Args} = format_log_single(Report, FormatOpts),
   io_lib:format(Format, Args, IoOpts).

format_log_single(#{label := {gen_event, terminate},
   handler := Handler,
   name := SName,
   last_message := LastIn,
   state := State,
   reason := Reason},
   #{single_line := true, depth := Depth} = FormatOpts) ->
   P = p(FormatOpts),
   Reason1 = fix_reason(Reason),
   Format1 = lists:append(["Generic event handler ", P, " crashed. "
   "Installed: ", P, ". Last event: ", P,
      ". State: ", P, ". Reason: ", P, "."]),
   Args1 =
      case Depth of
         unlimited ->
            [Handler, SName, Reason1, LastIn, State];
         _ ->
            [Handler, Depth, SName, Depth, Reason1, Depth,
               LastIn, Depth, State, Depth]
      end,
   {Format1, Args1};
format_log_single(#{label := {gen_event, no_handle_info},
   module := Mod,
   message := Msg},
   #{single_line := true, depth := Depth} = FormatOpts) ->
   P = p(FormatOpts),
   Format = lists:append(["Undefined handle_info in ", P,
      ". Unhandled message: ", P, "."]),
   Args =
      case Depth of
         unlimited ->
            [Mod, Msg];
         _ ->
            [Mod, Depth, Msg, Depth]
      end,
   {Format, Args};
format_log_single(Report, FormatOpts) ->
   format_log_multi(Report, FormatOpts).

format_log_multi(#{label := {gen_event, terminate},
   handler := Handler,
   name := SName,
   last_message := LastIn,
   state := State,
   reason := Reason},
   #{depth := Depth} = FormatOpts) ->
   Reason1 = fix_reason(Reason),
   P = p(FormatOpts),
   Format =
      lists:append(["** gen_event handler ", P, " crashed.\n",
         "** Was installed in ", P, "\n",
         "** Last event was: ", P, "\n",
         "** When handler state == ", P, "\n",
         "** Reason == ", P, "\n"]),
   Args =
      case Depth of
         unlimited ->
            [Handler, SName, LastIn, State, Reason1];
         _ ->
            [Handler, Depth, SName, Depth, LastIn, Depth, State, Depth,
               Reason1, Depth]
      end,
   {Format, Args};
format_log_multi(#{label := {gen_event, no_handle_info},
   module := Mod,
   message := Msg},
   #{depth := Depth} = FormatOpts) ->
   P = p(FormatOpts),
   Format =
      "** Undefined handle_info in ~p\n"
      "** Unhandled message: " ++ P ++ "\n",
   Args =
      case Depth of
         unlimited ->
            [Mod, Msg];
         _ ->
            [Mod, Msg, Depth]
      end,
   {Format, Args}.

fix_reason({'EXIT', {undef, [{M, F, A, _L} | _] = MFAs} = Reason}) ->
   case code:is_loaded(M) of
      false ->
         {'module could not be loaded', MFAs};
      _ ->
         case erlang:function_exported(M, F, length(A)) of
            true ->
               Reason;
            false ->
               {'function not exported', MFAs}
         end
   end;
fix_reason({'EXIT', Reason}) ->
   Reason;
fix_reason(Reason) ->
   Reason.

p(#{single_line := Single, depth := Depth, encoding := Enc}) ->
   "~" ++ single(Single) ++ mod(Enc) ++ p(Depth);
p(unlimited) ->
   "p";
p(_Depth) ->
   "P".

single(true) -> "0";
single(false) -> "".

mod(latin1) -> "";
mod(_) -> "t".

handler(Handler) when not Handler#handler.id ->
   Handler#handler.module;
handler(Handler) ->
   {Handler#handler.module, Handler#handler.id}.

the_handlers(MSL) ->
   [handler(Handler) || Handler <- MSL].

%% stop_handlers(MSL, ServerName) -> []

stop_handlers([Handler | T], SName) ->
   Mod = Handler#handler.module,
   do_terminate(Mod, Handler, stop, Handler#handler.state,
      stop, SName, shutdown),
   stop_handlers(T, SName);
stop_handlers([], _) ->
   [].

%% Message from the release_handler.
%% The list of modules got to be a set, i.e. no duplicate elements!
get_modules(MSL) ->
   Mods = [Handler#handler.module || Handler <- MSL],
   ordsets:to_list(ordsets:from_list(Mods)).

%%-----------------------------------------------------------------
%% Status information
%%-----------------------------------------------------------------
format_status(Opt, StatusData) ->
   [PDict, SysState, Parent, _Debug, [ServerName, _HibernateAfterTimeout, EpmHers, _Hib]] = StatusData,
   Header = gen:format_status_header("Status for gen_emm handler", ServerName),
   FmtMSL = [MS#handler{state = format_status(Opt, Mod, PDict, State)} || #handler{module = Mod, state = State} = MS <- MSL],
   [{header, Header}, {data, [{"Status", SysState}, {"Parent", Parent}]}, {items, {"Installed handlers", FmtMSL}}].

format_status(Opt, Mod, PDict, State) ->
   case erlang:function_exported(Mod, format_status, 2) of
      true ->
         Args = [PDict, State],
         case catch Mod:format_status(Opt, Args) of
            {'EXIT', _} -> State;
            Else -> Else
         end;
      false ->
         State
   end.

